"""quixote.config

Quixote configuration information.  This module provides both the
default configuration values, and some code that Quixote uses for
dealing with configuration info.  You should not edit the configuration
values in this file, since your edits will be lost if you upgrade to a
newer Quixote version in the future.  However, this is the canonical
source of information about Quixote configuration variables, and editing
the defaults here is harmless if you're just playing around and don't
care what happens in the future.
"""

__revision__ = "$Id: config.py,v 1.16 2000/12/08 22:46:10 gward Exp $"


# Note that the default values here generally prefer convenience and
# verbosity over security.  The assumption is that Quixote is more
# likely to be installed by web developers who want to play around with
# it than web administrators who just want it to go to work on day 1.
# Third-party Quixote applications should ship with more secure
# defaults, so that end-users can install Quixote and that application
# and not have to worry about security implications of the Quixote
# configuration.  If you want to be sure to run a tight ship, consider
# changing DISPLAY_EXCEPTIONS and SECURE_ERRORS.


# The root namespace is the Python namespace that will handle all
# requests to this Quixote application.  This is usually the name of a
# package or module, but could be a class instance or module object --
# any Python namespace!  (Allowing strings is a special case.)
ROOT_NAMESPACE = 'quixote.demo'

# Prefix of URLs that will be directed to Quixote.  This just tells
# Quixote where this application lives in your web site's URL-space;
# you'll also have to configure your web server so that requests to this
# URL go to the FCGI/CGI script for this Quixote application.
URL_PREFIX = '/q'
<<<<<<< config.py

# E-mail address to send errors to
# (None to send no mail at all)

#ERROR_EMAIL = None
ERROR_EMAIL = 'akuchlin@mems-exchange.org'
=======
>>>>>>> 1.16

<<<<<<< config.py
# Filename for logging debugging output (None for no log file.)
=======
# E-mail address to send errors to; None to send no mail at all.  This
# should probably be the email address of your web administrator.
ERROR_EMAIL = None
#ERROR_EMAIL = 'akuchlin@mems-exchange.org'

# Filename for logging debugging output; None for no debug log file.
DEBUG_LOG = None
#DEBUG_LOG = "/www/log/quixote-debug.log"
>>>>>>> 1.16

<<<<<<< config.py
#DEBUG_LOG = None
DEBUG_LOG = "/tmp/quixote-debug.log"

# Page shown when there's an access-control violation
=======
# Page shown when there's an access-control violation (relative to
# URL_PREFIX).
>>>>>>> 1.16
ACCESS_VIOLATION_PAGE = '/access_error'

# Filename for logging error messages (None for output to standard error,
# which should wind up in the Web servers log file.)

<<<<<<< config.py
#ERROR_LOG = None
ERROR_LOG = "/www/httpd/logs/quixote.log"
=======
# Filename for logging error messages; if None, error messages will be
# sent to standard error, so should wind up in the Web server's error
# log file.
ERROR_LOG = None
#ERROR_LOG = "/www/log/quixote-error.log"
>>>>>>> 1.16

# If true, exceptions will also be returned to the browser; otherwise,
# a generic error display will be used.
DISPLAY_EXCEPTIONS = 1

# If true, Quixote will service exactly one request at a time, and
# then exit.  This makes no difference when you're running as a
# straight CGI script, but it makes it easier to debug while running
# as a FastCGI script.
RUN_ONCE = 1

# If true, then any "resource not found" errors will result in a
# consistent, terse, mostly-useless message.  If false, then the
# exact cause of failure will be returned.
SECURE_ERRORS = 0

# Session-related variables
# =========================

# Domain to specify in setting cookies to hold the session ID
COOKIE_DOMAIN = ".example.com"

# Name and path of the cookie that will hold the ID
COOKIE_NAME = "Qsession"
COOKIE_PATH = "/"




__exports__ = ["index", "other"]

from pages import index
from quixote.errors import TraversalError

def other(request):
    return "Handled by a Python function."

def fact(n):
    f = n*1L
    while n > 1:
        f *= n
        n -= 1
    return f

class IntegerUI:

    __exports__ = ["index", "factorial"]

    def __init__(self, request, component):
        try:
            self.n = int(component)
        except ValueError, exc:
            raise TraversalError(request.get_path(), str(exc))
    
    def factorial(self, request):
        return fact(self.n)
        
    def index(self, request):
        return "You have selected the integer %s" % self.n

def __getname__(request, component):
    return IntegerUI(request, component)



"""quixote.errors

Exception classes used by Quixote
"""

class PublishError (Exception):
    """PublishError exceptions are raised due to some problem with the
    data provided by the client and are raised during the publishing
    process.  Quixote will abort the current request and return an error
    to the client.  The error returned to the client is controlled by the
    format method.

    The formatting done by the Quixote versions of these exceptions is
    very simple.  Applications will probably wish to raise application
    specific subclasses which do more sophisticated formatting.
    """

    def format (self, request):
        """format(request : HTTPRequest) -> string

        Called by Quixote to return a response to the user.
        """
        raise NotImplementedError


class TraversalError (PublishError):
    """Raised when a URL attempts to traverse to a bogus or non- existent
    Python namespace (from quixote.publish or from __getname__ methods).
    There are a number of reasons why this can happen, so these exceptions
    should have a clear and explicit message explaining the problem.
    However, that can reveal valuable information to attackers, so the
    'format()' method here will hide that useful information if the
    configuration variable SECURE_ERRORS is true.
    """

    # XXX it should be possible to make publish raise an application
    # specific subclass of this exception

    def __init__ (self, path, msg=None):
        self.path = path
        self.msg = msg # an optional, more detailed message

    def __str__ (self):
        if self.msg:
            return "%s: %s" % (self.path, self.msg)
        else:
            return self.path

    def format (self, request):
        from cgi import escape
        from quixote.config import config
        request.response.setStatus(404)
        page = "<html><body>%s</body></htm>" # this should go in a template
        if not config.secure_errors:
            msg = 'Resource "%s" not found: %s' % (self.path, self.msg)
        else:
            msg = 'Resource "%s" not found.' % self.path
        return page % escape(msg)


class QueryError (PublishError):
    """Should be raised raised if bad data provided in the query part of a
    URL or in the content of a POST request.  What constitutes bad data is
    solely application dependent (eg: letters in a form field when the
    application expects a number).

    The formatting done by this exception is very simple.  Applications
    will probably wish to raise subclasses which do more sophisticated
    formatting.
    """

    def __init__ (self, msg):
        self.msg = msg

    def format (self, request):
        request.response.setHeader("Content-Type", "text/plain")
        return self.msg


class AccessError (PublishError):
    """Should be raised raised if the client does not have access to the
    requested resource.  Usually applications will raise this error from
    an __access__ method.
    """

    def __init__ (self, msg=None):
        self.msg = msg or "access denied"
    
    def __str__ (self):
        return self.msg

    def format (self, request):
        request.response.setHeader("Content-Type", "text/plain")
        request.response.setStatus(403) # Forbidden, is this the right status?
        return self.msg
    




##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################
'''CGI Response Output formatter

$Id: HTTPResponse.py,v 1.10 2001/01/10 22:22:45 nascheme Exp $'''
__version__='$Revision: 1.10 $'[11:-2]

import string, types, sys, regex
from string import find, rfind, lower, upper, strip, split, join, translate
from types import StringType, InstanceType
from BaseResponse import BaseResponse

nl2sp=string.maketrans('\n',' ')

status_reasons={
100: 'Continue',
101: 'Switching Protocols',
102: 'Processing',
200: 'OK',
201: 'Created',
202: 'Accepted',
203: 'Non-Authoritative Information',
204: 'No Content',
205: 'Reset Content',
206: 'Partial Content',
207: 'Multi-Status',
300: 'Multiple Choices',
301: 'Moved Permanently',
302: 'Moved Temporarily',
303: 'See Other',
304: 'Not Modified',
305: 'Use Proxy',
307: 'Temporary Redirect',
400: 'Bad Request',
401: 'Unauthorized',
402: 'Payment Required',
403: 'Forbidden',
404: 'Not Found',
405: 'Method Not Allowed',
406: 'Not Acceptable',
407: 'Proxy Authentication Required',
408: 'Request Time-out',
409: 'Conflict',
410: 'Gone',
411: 'Length Required',
412: 'Precondition Failed',
413: 'Request Entity Too Large',
414: 'Request-URI Too Large',
415: 'Unsupported Media Type',
416: 'Requested range not satisfiable',
417: 'Expectation Failed',
422: 'Unprocessable Entity',
423: 'Locked',
424: 'Failed Dependency',
500: 'Internal Server Error',
501: 'Not Implemented',
502: 'Bad Gateway',
503: 'Service Unavailable',
504: 'Gateway Time-out',
505: 'HTTP Version not supported',
507: 'Insufficient Storage',
}

status_codes={}
# Add mappings for builtin exceptions and
# provide text -> error code lookups.
for key, val in status_reasons.items():
    status_codes[lower(join(split(val, ' '), ''))]=key
    status_codes[lower(val)]=key
    status_codes[key]=key
    status_codes[str(key)]=key
en=filter(lambda n: n[-5:]=='Error', dir(__builtins__))
for name in map(lower, en):
    status_codes[name]=500
status_codes['nameerror']=503
status_codes['keyerror']=503
status_codes['redirect']=300


accumulate_header={'set-cookie': 1}.has_key

class HTTPResponse(BaseResponse):
    """\
    An object representation of an HTTP response.
    
    The Response type encapsulates all possible responses to HTTP
    requests.  Responses are normally created by the object publisher.
    A published object may recieve the response abject as an argument
    named 'RESPONSE'.  A published object may also create it's own
    response object.  Normally, published objects use response objects
    to:

    - Provide specific control over output headers,

    - Set cookies, or

    - Provide stream-oriented output.

    If stream oriented output is used, then the response object
    passed into the object must be used.
    """ #'

    accumulated_headers=''
    body=''

    def __init__(self,body='', status=200, headers=None,
                 stdout=sys.stdout, stderr=sys.stderr):
        '''\
        Creates a new response. In effect, the constructor calls
        "self.setBody(body); self.setStatus(status); for name in
        headers.keys(): self.setHeader(name, headers[name])"
        '''
        if headers is None: headers={}
        self.headers=headers
        
        if status==200:
            self.status=200
            self.errmsg='OK'
            headers['status']="200 OK"      
        else: self.setStatus(status)
        self.base=''
        if body: self.setBody(body)
        self.cookies={}
        self.stdout=stdout
        self.stderr=stderr

    def setStatus(self, status, reason=None):
        '''\
        Sets the HTTP status code of the response; the argument may
        either be an integer or a string from { OK, Created, Accepted,
        NoContent, MovedPermanently, MovedTemporarily,
        NotModified, BadRequest, Unauthorized, Forbidden,
        NotFound, InternalError, NotImplemented, BadGateway,
        ServiceUnavailable } that will be converted to the correct
        integer value. '''
        if type(status) is types.StringType:
            status=lower(status)
        if status_codes.has_key(status): status=status_codes[status]
        else: status=500
        self.status=status
        if reason is None:
            if status_reasons.has_key(status): reason=status_reasons[status]
            else: reason='Unknown'
        self.setHeader('Status', "%d %s" % (status,str(reason)))
        self.errmsg=reason

    def setHeader(self, name, value, literal=0):
        '''\
        Sets an HTTP return header "name" with value "value", clearing
        the previous value set for the header, if one exists. If the
        literal flag is true, the case of the header name is preserved,
        otherwise word-capitalization will be performed on the header
        name on output.'''
        key=lower(name)
        if accumulate_header(key):
            self.accumulated_headers=(
                "%s%s: %s\n" % (self.accumulated_headers, name, value))
            return
        name=literal and name or key
        self.headers[name]=value

    def addHeader(self, name, value):
        '''\
        Set a new HTTP return header with the given value, while retaining
        any previously set headers with the same name.'''
        self.accumulated_headers=(
            "%s%s: %s\n" % (self.accumulated_headers, name, value))

    __setitem__=setHeader

    def setBody(self, body):
        '''Set the body of the response
        
        Sets the return body equal to the (string) argument "body". Also
        updates the "Content-length" return header.  May also set the
        "Content-type" header if it hasn't already been set.
        '''
        if not body: return self
        
        if type(body) is not types.StringType:
            if hasattr(body,'asHTML'):
                body=body.asHTML()

        self.body=str(body)
        self.setHeader('content-length',len(self.body))
        if not self.headers.has_key('content-type'):
            self.setHeader('content-type','text/html')
        return self

    def appendCookie(self, name, value):
        '''\
        Returns an HTTP header that sets a cookie on cookie-enabled
        browsers with a key "name" and value "value". If a value for the
        cookie has previously been set in the response object, the new
        value is appended to the old one separated by a colon. '''

        cookies=self.cookies
        if cookies.has_key(name): cookie=cookies[name]
        else: cookie=cookies[name]={}
        if cookie.has_key('value'):
            cookie['value']='%s:%s' % (cookie['value'], value)
        else: cookie['value']=value

    def expireCookie(self, name, **kw):
        '''\
        Cause an HTTP cookie to be removed from the browser
        
        The response will include an HTTP header that will remove the cookie
        corresponding to "name" on the client, if one exists. This is
        accomplished by sending a new cookie with an expiration date
        that has already passed. Note that some clients require a path
        to be specified - this path must exactly match the path given
        when creating the cookie. The path can be specified as a keyword
        argument.
        '''
        dict={'max_age':0, 'expires':'Wed, 31-Dec-97 23:59:59 GMT'}
        for k, v in kw.items():
            dict[k]=v
        apply(HTTPResponse.setCookie, (self, name, 'deleted'), dict)

    def setCookie(self,name,value,**kw):
        '''\
        Set an HTTP cookie on the browser

        The response will include an HTTP header that sets a cookie on
        cookie-enabled browsers with a key "name" and value
        "value". This overwrites any previously set value for the
        cookie in the Response object.
        '''
        cookies=self.cookies
        if cookies.has_key(name):
            cookie=cookies[name]
        else: cookie=cookies[name]={}
        for k, v in kw.items():
            cookie[k]=v
        cookie['value']=value

    def appendHeader(self, name, value, delimiter=","):
        '''\
        Append a value to a cookie
        
        Sets an HTTP return header "name" with value "value",
        appending it following a comma if there was a previous value
        set for the header. '''
        headers=self.headers
        if headers.has_key(name):
            h=self.header[name]
            h="%s%s\n\t%s" % (h,delimiter,value)
        else: h=value
        self.setHeader(name,h)

    def quoteHTML(self,text,
                  subs={'&':'&amp;', "<":'&lt;', ">":'&gt;', '\"':'&quot;'}
                  ):
        for ent in '&<>\"':
            if find(text, ent) >= 0:
                text=join(split(text,ent),subs[ent])

        return text
         

    def redirect(self, location, permanent=0):
        """Cause a redirection without raising an error"""
        # Ensure that location is a full URL
        if string.find(location, '://') == -1:
            raise ValueError, "URL must include the server name"
        if permanent:
            self.status = 301
        else:
            self.status = 302
        self.headers['status'] = "%d %s" % (self.status,
                                            status_reasons[self.status])
        self.headers['location'] = location

        doc = """\
<html>
 <head><title>Redirect to %(location)s</title></head>
 <body>
 <h1>Redirect</h1>
 <p>Your browser should have redirected you to
 <center><a href="%(location)s">%(location)s</a></center>
 </p>
 </body>
</html>
""" % vars()

        return doc


    def _cookie_list(self):
        cookie_list=[]
        for name, attrs in self.cookies.items():

            # Note that as of May 98, IE4 ignores cookies with
            # quoted cookie attr values, so only the value part
            # of name=value pairs may be quoted.

            cookie='Set-Cookie: %s="%s"' % (name, attrs['value'])
            for name, v in attrs.items():
                name=lower(name)
                if name=='expires': cookie = '%s; Expires=%s' % (cookie,v)
                elif name=='domain': cookie = '%s; Domain=%s' % (cookie,v)
                elif name=='path': cookie = '%s; Path=%s' % (cookie,v)
                elif name=='max_age': cookie = '%s; Max-Age=%s' % (cookie,v)
                elif name=='comment': cookie = '%s; Comment=%s' % (cookie,v)
                elif name=='secure' and v: cookie = '%s; Secure' % cookie
            cookie_list.append(cookie)

        # Should really check size of cookies here!
        
        return cookie_list

    _wrote=None

    def __str__(self):
        if self._wrote: return ''       # Streaming output was used.

        headers=self.headers
        headersl=[]
        append=headersl.append

        # status header must come first.
        append("Status: %s" % headers.get('status', '200 OK'))
        for key, val in headers.items():
            # Skip the status header since it will have already been output
            if key == 'status': continue
            if lower(key)==key:
                # only change non-literal header names
                key="%s%s" % (upper(key[:1]), key[1:])
                start=0
                l=find(key,'-',start)
                while l >= start:
                    key="%s-%s%s" % (key[:l],upper(key[l+1:l+2]),key[l+2:])
                    start=l+1
                    l=find(key,'-',start)
            append("%s: %s" % (key, val))
        if self.cookies:
            headersl=headersl+self._cookie_list()
        headersl[len(headersl):]=[self.accumulated_headers, self.body]
        return join(headersl,'\n')

    def write(self,data):
        """Return data as a stream

        HTML data may be returned using a stream-oriented interface.
        This allows the browser to display partial results while
        computation of a response to proceed.

        The published object should first set any output headers or
        cookies on the response object.

        Note that published objects must not generate any errors
        after beginning stream-oriented output. 

        """
        if not self._wrote:
            self.outputBody()
            self._wrote=1
            self.stdout.flush()

        self.stdout.write(data)



##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################

__version__='$Revision: 1.17 $'[11:-2]

import re, string
from string import lower, rfind, split, strip, join, upper, find
from BaseRequest import BaseRequest
from HTTPResponse import HTTPResponse
from cgi import FieldStorage
from urllib import quote, unquote
from types import ListType

isCGI_NAME = {
        'SERVER_SOFTWARE' : 1, 
        'SERVER_NAME' : 1, 
        'GATEWAY_INTERFACE' : 1, 
        'SERVER_PROTOCOL' : 1, 
        'SERVER_PORT' : 1, 
        'REQUEST_METHOD' : 1, 
        'PATH_INFO' : 1, 
        'PATH_TRANSLATED' : 1, 
        'SCRIPT_NAME' : 1, 
        'QUERY_STRING' : 1, 
        'REMOTE_HOST' : 1, 
        'REMOTE_ADDR' : 1, 
        'AUTH_TYPE' : 1, 
        'REMOTE_USER' : 1, 
        'REMOTE_IDENT' : 1, 
        'CONTENT_TYPE' : 1, 
        'CONTENT_LENGTH' : 1,
        'SERVER_URL': 1,
        }.has_key

hide_key={'HTTP_AUTHORIZATION':1,
          'HTTP_CGI_AUTHORIZATION': 1,
          }.has_key


class HTTPRequest(BaseRequest):
    """\
    Model HTTP request data.
    
    This object provides access to request data.  This includes, the
    input headers, form data, server data, and cookies.

    Request objects are created by the object publisher and will be
    passed to published objects through the argument name, REQUEST.

    The request object is a mapping object that represents a
    collection of variable to value mappings.  In addition, variables
    are divided into four categories:

      - Environment variables

        These variables include input headers, server data, and other
        request-related data.  The variable names are as <a
        href="http://hoohoo.ncsa.uiuc.edu/cgi/env.html">specified</a>
        in the <a
        href="http://hoohoo.ncsa.uiuc.edu/cgi/interface.html">CGI
        specification</a>

      - Form data

        These are data extracted from either a URL-encoded query
        string or body, if present.

      - Cookies

        These are the cookie data, if present.

      - Other

        Data that may be set by an application object.

    The form attribute of a request is actually a Field Storage
    object.  When file uploads are used, this provides a richer and
    more complex interface than is provided by accessing form data as
    items of the request.  See the FieldStorage class documentation
    for more details.

    The request object may be used as a mapping object, in which case
    values will be looked up in the order: environment variables,
    form data, and then cookies.
    """
    _file=None

    def __init__(self, stdin, environ):
        self.stdin = stdin
        self.environ = environ
        self.form = {}
        self.session = None
        self.response = HTTPResponse()

        if environ.has_key('HTTP_AUTHORIZATION'):
            self._auth=environ['HTTP_AUTHORIZATION']
            response._auth=1
            del environ['HTTP_AUTHORIZATION']

        if (lower(environ.get('HTTPS', "off")) == "on" or
                environ.get('SERVER_PORT_SECURE', 0)):
            self.scheme = "https"
        else:
            self.scheme = "http"

        ################################################################
        # Cookie values should *not* be appended to existing form
        # vars with the same name - they are more like default values
        # for names not otherwise specified in the form.
        cookies={}
        k = self.environ.get('HTTP_COOKIE','')
        if k:
            parse_cookie(k, cookies)
        self.cookies=cookies

        
    def processInputs(self):
        """Process request inputs

        We need to delay input parsing so that it is done under
        publisher control for error handling purposes.
        """
        environ=self.environ
        method = self.environ.get('REQUEST_METHOD','GET')
        
        if method != 'GET':
            fp=self.stdin
        else:
            fp=None

        fs=FieldStorage(fp=fp, environ=self.environ, keep_blank_values=1)
        if not hasattr(fs,'list') or fs.list is None:
            self._file=fs.file
        else:
            for item in fs.list:
                key=item.name
                if (hasattr(item,'file') and hasattr(item,'filename')
                    and hasattr(item,'headers')):
                    if (item.file and
                        (item.filename is not None
                         # RFC 1867 says that all fields get a content-type.
                         # or 'content-type' in map(lower, item.headers.keys())
                         )):
                        item=FileUpload(item)
                    else:
                        item=item.value

                # Filter out special names from form:
                if isCGI_NAME(key) or key[:5]=='HTTP_':
                    continue

                #Insert in dictionary
                if self.form.has_key(key):
                    # it is not a record or list of records
                    found=self.form[key]
                    if type(found) is ListType:
                        found.append(item)
                    else:
                        found=[found,item]
                        self.form[key]=found
                else:
                    self.form[key]=item
                                
    def get_header(self, name, default=None):
        """Return the named HTTP header, or an optional default
        argument or None if the header is not found. Note that
        both original and CGI-ified header names are recognized,
        e.g. 'Content-Type', 'CONTENT_TYPE' and 'HTTP_CONTENT_TYPE'
        should all return the Content-Type header, if available.
        """
        environ=self.environ
        name=upper(join(split(name,"-"),"_"))
        val=environ.get(name, None)
        if val is not None:
            return val
        if name[:5] != 'HTTP_':
            name='HTTP_%s' % name
        return environ.get(name, default)


    # The following environmental variables are useful for reconstructing
    # the original URL, all of which are specified by CGI 1.1:
    #
    #   SERVER_NAME            "www.example.com"
    #   SCRIPT_NAME            "/q"
    #   PATH_INFO              "/debug/dump_sessions"
    #   QUERY_STRING           "session_id=10.27.8.40...."

    def get_server(self):
        """Return the server name with an optional port number."""

        server_name = strip(self.environ["SERVER_NAME"])
        server_port = self.environ.get("SERVER_PORT")

        if (not server_port or
                (self.scheme == "http" and server_port == "80") or
                (self.scheme == "https" and server_port == "443")):
            return server_name
        else:
            return server_name + ":" + server_port

    def get_path(self, n=0):
        """Return the path of the current request, chopping off 'n' path
        components from the right.  Eg. if the path is "/bar/baz/qux", n=0
        would return "/bar/baz/qux" and n=2 would return "/bar".  Note
        that the query string is not included.  Also note the handling of
        a path with a trailing slash.  If the original path is "/bar/baz/"
        then get_path(0) == "/bar/baz/" and get_path(1) == "/bar/baz/"
        (ie. the last component of a path with a trailing slash is the
         empty string).
        """
        path = "%(SCRIPT_NAME)s%(PATH_INFO)s" % self.environ
        if n == 0:
            return path
        else:
            path_comps = string.split(path, '/')
            if n > len(path_comps)-1:
                raise ValueError, "n=%d too big for path '%s'" % (n, path)
            return string.join(path_comps[:-n], '/')

    def get_url(self, n=0):
        """Return the URL of the current request, chopping off 'n' path
        components from the right.  Eg. if the URL is
        "http://foo.com/bar/baz/qux", n=2 would return
        "http://foo.com/bar".  Does not include the query string (if any).
        """
        return "%s://%s%s" % (self.scheme, self.get_server(), self.get_path(n))

    def get_environ (self, key, default=None):
        """get_environ(key : string) -> string

        Fetch a CGI environment variable from the request environment.
        See http://hoohoo.ncsa.uiuc.edu/cgi/env.html
        for the variables specified by the CGI standard.
        """
        return self.environ.get(key, default)

    def get(self, key, default=None):
        if self.environ.has_key(key):
            return self.environ[key]
        elif self.form.has_key(key):
            return self.form[key]
        else:
            return default

    def dump_html(self):
        result="<h3>form</h3><table>"
        row='<tr valign="top" align="left"><th>%s</th><td>%s</td></tr>'
        for k,v in self.form.items():
            result=result + row % (html_quote(k), html_quote(v))
        result=result+"</table><h3>cookies</h3><table>"
        for k,v in self.cookies.items():
            result=result + row % (html_quote(k), html_quote(v))
    
        result=result+"</table><h3>environ</h3><table>"
        for k,v in self.environ.items():
            if not hide_key(k):
                result=result + row % (html_quote(k), html_quote(v))
        return result+"</table>"

    def dump(self):
        result=""
        row='%-15s %s\n'
        L = self.form.items() ; L.sort()
        for k,v in L:
            result=result + row % (k,v)
        result=result+"\nCookies:\n"
        L = self.cookies.items() ; L.sort()
        for k,v in L:
            result=result + row % (k,v)

        result=result+"\nEnvironment:\n"
        L = self.environ.items() ; L.sort()
        for k,v in L:
            if not hide_key(k):
                result=result + row % (k,v)
        return result

    def browser_version (self):
        """Examine the User-agent request header to try to figure out
           what the current browser is.  Returns a (name,version) tuple
           where each element is a string or None if we couldn't parse the
           User-agent header.  Note that Netscape {Communicator,Navigator}
           and Internet Explorer advertise themselves as, respectively,
           "Mozilla" and "MSIE"."""

        ua = self.get_header ('user-agent')
        if ua is None:
            return (None, None)

        # Looks like Netscape...
        moz_match = re.search (r'(Mozilla)/([\w\.]+)', ua)
        if moz_match:
            # ...but smells like Microsoft
            ie_match = re.search (r'(MSIE)\s+([\w\.]+)', ua)
            if ie_match:
                return ie_match.group (1, 2)

            # nope, it's really Netscape
            else:
                return moz_match.group (1, 2)

        # neither Netscape or IE -- try a broader regex
        else:
            match = re.search (r'([^\s\/]+)/([\w\.]+)', ua)
            if match:
                return match.group (1, 2)
            else:
                return (None, None)

    # browser_version ()
    
    def _authUserPW(self):
        global base64
        auth=self._auth
        if auth:
            if lower(auth[:6]) == 'basic ':
                if base64 is None: import base64
                [name,password] = split(
                    base64.decodestring(split(auth)[-1]), ':')
                return name, password

    def redirect(self, location, permanent=0):
        """Create a redirection response.  If the location does not
        specify the server then it is automatically added.  The return
        value is a HTML document indicating the new URL (useful if the
        client browser does not honor the redirect).
        """
        if string.find(location, "://") == -1:
            location = "%s://%s%s" % (self.scheme, self.get_server(), location)
        return self.response.redirect(location, permanent)


base64=None


# This is duplicated from DocumentTemplate.DT_Util to
# prevent a dependency on the DocumentTemplate package.
# Some folks still use the ZPublisher package as a
# standalone publisher without DocumentTemplate.
def html_quote(value, character_entities=(
                      (('&'), '&amp;'),
                      (("<"), '&lt;' ),
                      ((">"), '&gt;' ),
                      (('"'), '&quot;'))): #"
        text=str(value)
        for re, name in character_entities:
            if find(text, re) >= 0: text=join(split(text, re), name)
        return text


def str_field(v):
    if type(v) is ListType:
        return map(str_field,v)

    if hasattr(v,'__class__') and v.__class__ is FieldStorage:
        v=v.value
    elif type(v) is not StringType:
        if hasattr(v,'file') and v.file: v=v.file
        elif hasattr(v,'value'): v=v.value
    return v


class FileUpload:
    '''\
    File upload objects

    File upload objects are used to represent file-uploaded data.

    File upload objects can be used just like files.

    In addition, they have a 'headers' attribute that is a dictionary
    containing the file-upload headers, and a 'filename' attribute
    containing the name of the uploaded file.
    '''

    def __init__(self, aFieldStorage):

        file=aFieldStorage.file
        if hasattr(file, '__methods__'): methods=file.__methods__
        else: methods= ['close', 'fileno', 'flush', 'isatty',
                        'read', 'readline', 'readlines', 'seek',
                        'tell', 'truncate', 'write', 'writelines']

        d=self.__dict__
        for m in methods:
            if hasattr(file,m): d[m]=getattr(file,m)

        self.headers=aFieldStorage.headers
        self.filename=aFieldStorage.filename
    

# XXX this should be converted to use re
import regex

def parse_cookie(text,
                 result=None,
                 qparmre=regex.compile(
                     '\([\0- ]*'
                     '\([^\0- ;,=\"]+\)="\([^"]*\)\"'
                     '\([\0- ]*[;,]\)?[\0- ]*\)'
                     ),
                 parmre=regex.compile(
                     '\([\0- ]*'
                     '\([^\0- ;,=\"]+\)=\([^\0- ;,\"]*\)'
                     '\([\0- ]*[;,]\)?[\0- ]*\)'
                     ),
                 ):

    if result is None: result={}
    already_have=result.has_key

    try:
        if qparmre.match(text) >= 0:
            # Match quoted correct cookies
            name=qparmre.group(2)
            value=qparmre.group(3)
            l=len(qparmre.group(1))
        elif parmre.match(text) >= 0:
            # Match evil MSIE cookies ;)
            name=parmre.group(2)
            value=parmre.group(3)
            l=len(parmre.group(1))
        else:
            # this may be an invalid cookie.
            # We'll simply bail without raising an error
            # if the cookie is invalid.
            return result
            
    finally: pass

    if not already_have(name): result[name]=value

    return apply(parse_cookie,(text[l:],result))

# add class
class record:
    def __getattr__(self, key, default=None):
        if key in ('get', 'keys', 'items', 'values', 'copy', 'has_key'):
            return getattr(self.__dict__, key)
        raise AttributeError, key

    def __getitem__(self, key):
        return self.__dict__[key]
            
    def __str__(self):
        L1 = self.__dict__.items()
        L1.sort()
        return join(map(lambda item: "%s: %s" %item, L1), ", ") 

    __repr__ = __str__




##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################
'''CGI Response Output formatter

$Id: BaseResponse.py,v 1.3 2000/10/31 16:27:01 gward Exp $'''
__version__='$Revision: 1.3 $'[11:-2]

import string, types, sys, regex
from string import find, rfind, lower, upper, strip, split, join, translate
from types import StringType, InstanceType

class BaseResponse:
    """Base Response Class

    What should be here?
    """
    debug_mode=None
    _auth=None
    _error_format='text/plain'

    def __init__(self, stdout, stderr,
                 body='', headers=None, status=None, cookies=None):
        self.stdout=stdout
        self.stderr=stderr
        self.body=body
        if headers is None: headers={}
        self.headers=headers
        self.status=status
        if cookies is None: cookies={}
        self.cookies=cookies
    
    def setStatus(self, status, reason=None):
        self.status=status

    def setHeader(self, name, value):
        self.headers[n]=value

    __setitem__=setHeader

    def outputBody(self):
        """Output the response body"""
        self.stdout.write(str(self))

    def setBody(self, body):
        self.body=body

    def getStatus(self):
        'Returns the current HTTP status code as an integer. '
        return self.status

    def setCookie(self,name,value,**kw):
        '''\
        Set an HTTP cookie on the browser

        The response will include an HTTP header that sets a cookie on
        cookie-enabled browsers with a key "name" and value
        "value". This overwrites any previously set value for the
        cookie in the Response object.
        '''
        cookies=self.cookies
        if cookies.has_key(name):
            cookie=cookies[name]
        else: cookie=cookies[name]={}
        for k, v in kw.items():
            cookie[k]=v
        cookie['value']=value

    def appendBody(self, body):
        self.setBody(self.getBody() + body)

    def getHeader(self, name):
         '''\
         Get a header value
         
         Returns the value associated with a HTTP return header, or
         "None" if no such header has been set in the response
         yet. '''
         return self.headers.get(name, None)

    def __getitem__(self, name):
        'Get the value of an output header'
        return self.headers[name]

    def getBody(self):
        'Returns a string representing the currently set body. '
        return self.body

    def flush(self): pass

    def write(self,data):
        """\
        Return data as a stream

        HTML data may be returned using a stream-oriented interface.
        This allows the browser to display partial results while
        computation of a response to proceed.

        The published object should first set any output headers or
        cookies on the response object.

        Note that published objects must not generate any errors
        after beginning stream-oriented output. 

        """
        self.body=self.body+data




"""quixote.init

Stuff for setting up Quixote prior to entering the main loop.

Note that if you're using ZODB, you must import ZODB before installing
Quixote's import hooks -- we're not sure why this is, but it is.  This
applies to 'init_quixote()' and 'install_imphooks()'.
"""

# created 2000/08/01, GPW

__revision__ = "$Id: init.py,v 1.6 2000/12/11 16:58:07 gward Exp $"

import sys
from types import StringType

def init_quixote ():
    install_imphooks()
    setup_logs()


def install_imphooks ():
    from quixote import imphooks
    imphooks.install()


def read_config_file (filename):
    from quixote.config import config
    config.read_file(filename)

def set_config (config):
    import quixote.config
    if not isinstance(config, quixote.config.Config):
        raise TypeError, "'config' must be a Config instance"
    quixote.config.config = config


class AutoflushFile:
    """A file-like object (output only) that flushes on every write."""

    def __init__ (self, file, mode="w"):
        if type(file) is StringType:
            file = open(file, mode)
        self.file = file

    def close (self):
        self.file.close()

    def write (self, str):
        self.file.write(str)
        self.file.flush()

    def writelines (self, list):
        self.file.writelines(list)
        self.file.flush()

    # Unnecessary due to the nature of the beast, but here for
    # a more complete implementation of the "file" interface.
    def flush (self):
        pass


_orig_stdout = None

def setup_logs ():
    """Redirect stdout: either to /dev/null or a debug log file
    (depending on whether DEBUG_LOG is defined in config.py).
    Saves the original stdout so it can be restored by 'restore_stdout()'.
    Note that this is not a real redirect in that it doesn't affect
    child processes; this just reassigns 'sys.stdout'.
    """
    from quixote.config import config
    global _orig_stdout

    if _orig_stdout is not None:
        raise RuntimeError, "you've already redirected stdout!"
    _orig_stdout = sys.stdout

    if config.debug_log is None:
        # If no debug log file is configured, debugging output goes
        # to the bit-bucket.
        sys.stdout = open('/dev/null', 'w')
    else:
        sys.stdout = AutoflushFile(config.debug_log, 'a')

def get_orig_stdout ():
    global _orig_stdout
    if _orig_stdout is None:
        raise RuntimeError, "redirect_stdout never called"
    return _orig_stdout

def shutdown_logs ():
    """Restore 'sys.stdout' to what it was before calling
    'setup_logs()'.
    """
    global _orig_stdout
    if _orig_stdout is None:
        raise RuntimeError, "'setup_logs()' never called"
    sys.stdout.close()
    sys.stdout = _orig_stdout

# aliases for backwards compatibility
redirect_stdout = setup_logs
restore_stdout = shutdown_logs        



#!/usr/bin/env python

# Setup script for Quixote

from distutils.core import setup

setup (name = "Quixote",
       version = "0.2",
       description = "Quixote Web application development system",
       author = "MEMS Exchange",
       author_email = "quixote@mems-exchange.org",
       url = "http://www.mems-exchange.org/software/python/quixote/",

       package_dir = {'quixote':'.'},
       packages = ['quixote',  'quixote.demo', 'quixote.test',
                   'quixote.util', 'quixote.zope'],
      )

 



"""quixote.config

Quixote configuration information.  This module provides both the
default configuration values, and some code that Quixote uses for
dealing with configuration info.  You should not edit the configuration
values in this file, since your edits will be lost if you upgrade to a
newer Quixote version in the future.  However, this is the canonical
source of information about Quixote configuration variables, and editing
the defaults here is harmless if you're just playing around and don't
care what happens in the future.
"""

__revision__ = "$Id: config.py,v 1.20 2001/01/10 22:23:07 nascheme Exp $"


# Note that the default values here generally prefer convenience and
# verbosity over security.  The assumption is that Quixote is more
# likely to be installed by web developers who want to play around with
# it than web administrators who just want it to go to work on day 1.
# Third-party Quixote applications should ship with more secure
# defaults, so that end-users can install Quixote and that application
# and not have to worry about security implications of the Quixote
# configuration.  If you want to be sure to run a tight ship, consider
# changing DISPLAY_EXCEPTIONS and SECURE_ERRORS.


# The root namespace is the Python namespace that will handle all
# requests to this Quixote application.  This is usually the name of a
# package or module, but could be a class instance or module object --
# any Python namespace!  (Allowing strings is a special case.)
ROOT_NAMESPACE = 'quixote.demo'

# Prefix of URLs that will be directed to Quixote.  This just tells
# Quixote where this application lives in your web site's URL-space;
# you'll also have to configure your web server so that requests to this
# URL go to the FCGI/CGI script for this Quixote application.
URL_PREFIX = '/q'

# E-mail address to send errors to; None to send no mail at all.  This
# should probably be the email address of your web administrator.
ERROR_EMAIL = None
#ERROR_EMAIL = 'akuchlin@mems-exchange.org'

# Filename for logging debugging output; None for no debug log file.
DEBUG_LOG = None
#DEBUG_LOG = "/www/log/quixote-debug.log"

# Function called when there is an unhandled exception.  The signature of
# the function needs to be:
#
#   error_handler(request, exception, error_message) -> string".
#
# If this is set to None Quixote's default error handler is used.
ERROR_HANDLER = None

# Filename for logging error messages; if None, error messages will be
# sent to standard error, so should wind up in the Web server's error
# log file.
ERROR_LOG = None

# If true, exceptions will also be returned to the browser; otherwise, a
# generic error display will be used.  It is convenient to enable this on
# development servers.  On publicly accessible servers it should be
# disabled for security reasons.
DISPLAY_EXCEPTIONS = 0

# If true, then any "resource not found" errors will result in a
# consistent, terse, mostly-useless message.  If false, then the
# exact cause of failure will be returned.
SECURE_ERRORS = 1

# If true, Quixote will service exactly one request at a time, and
# then exit.  This makes no difference when you're running as a
# straight CGI script, but it makes it easier to debug while running
# as a FastCGI script.
RUN_ONCE = 0

# Automaticly redirect paths referencing non-callable objects to a path
# with a trailing slash.  This is convienent for external users of the
# site and but should be disabled for development.  Internal links on the
# site should not require redirects.  They are costly, especially on high
# latency links like dialup lines.
FIX_TRAILING_SLASH = 1


# Session-related variables
# =========================

# Domain to specify in setting cookies to hold the session ID
COOKIE_DOMAIN = ".example.com"

# Name and path of the cookie that will hold the ID
COOKIE_NAME = "Qsession"
COOKIE_PATH = "/"


# -- End config variables ----------------------------------------------
# (no user serviceable parts after this point)

# Note that this module is designed to not export any names apart from
# the above config variables and the following Config class -- hence,
# all imports are done in local scopes.  This allows application config
# modules to safely say "from quixote.config import *".

class Config:
    """Holds all Quixote configuration variables -- see above for
    documentation of them.  The naming convention is simple:
    downcase the above variables to get the names of instance
    attributes of this class.
    """

    config_vars = [
        'root_namespace',
        'url_prefix',
        'error_email',
        'debug_log',
        'display_exceptions',
        'secure_errors',
        'error_log',
        'error_handler',
        'run_once',
        'fix_trailing_slash',
        'cookie_domain',
        'cookie_name',
        'cookie_path',
        ]


    def __init__ (self):
        for var in self.config_vars:
            setattr(self, var, None)

    def __setattr__ (self, attr, val):
        if not attr in self.config_vars:
            raise AttributeError, "no such configuration variable: %s" % `attr`
        self.__dict__[attr] = val


    def dump (self, file=None):
        import sys
        if file is None:
            file = sys.stdout
        file.write("<%s.%s instance at %x>:\n" %
                   (self.__class__.__module__,
                    self.__class__.__name__,
                    id(self)))
        for var in self.config_vars:
            file.write("  %s = %s\n" % (var, `getattr(self, var)`))


    def _set_from_dict (self, dict):
        import string, re
        ucstring_re = re.compile(r'^[A-Z_]+$')

        for (var, val) in dict.items():
            if ucstring_re.match(var):
                setattr(self, string.lower(var), val)

    def read_file (self, filename):
        """Read configuration from a file.  Any variables already
        defined in this Config instance, but not in the file, are  
        unchanged, so you can use this to build up a configuration  
        by accumulating data from several config files.
        """
        # The config file is Python code -- makes life easy.
        config_vars = {}
        try:
            execfile(filename, config_vars)
        except IOError, exc:
            if exc.filename is None:    # arg! execfile() loses filename
                exc.filename = filename
            raise exc

        self._set_from_dict(config_vars)

    def read_from_module (self, modname):
        """Read configuration info from a Python module (default
        is the module where the Config class is defined, ie.
        quixote.config).  Also accumulates config data, just like
        'read_file()'.
        """
        import sys
        __import__(modname)
        module = sys.modules[modname]
        self._set_from_dict(vars(module))

    def read_default (self):
        self.read_from_module("quixote.config")

# class Config


# The default config object.  This is initialized with all the default
# values from the config variables at the top of this file, ie. if
# ROOT_NAMESPACE is set to "quixote.demo", then config.root_namespace
# will also be "quixote.demo".  If application FCGI/CGI scripts need to
# override any of these defaults, they can do so by direct manipulation
# of this config object, or by reading a config file:
#   from quixote.config import config
#   config.read_file("myapp.conf")
# or, equivalently,
#   from quixote.init import read_config_file
#   read_config_file("myapp.conf")

config = Config()
config.read_default()



"""quixote.publish

Logic for publishing modules and objects on the Web.
"""

__revision__ = "$Id: publish.py,v 1.55 2001/01/11 15:34:31 nascheme Exp $"

import sys, traceback, os, string, cStringIO
import time, types, socket, re

from quixote import init, session, errors
from quixote.util import fcgi
from quixote.zope import HTTPRequest, HTTPResponse


def get_module(name):
    """Get a module object by name."""
    __import__(name)
    module = sys.modules[name]
    return module
    
def _get_module_filename(module):
    filename = module.__file__
    if os.path.isdir( filename ):
        filename = os.path.join( filename, '__init__.py')
    elif filename[ -4: ] == '.pyc': filename = filename[:-1]

    return filename


def get_component (container, component, path, request):
    # First security check: if the container doesn't even have an
    # __exports__ list, fail now: all Quixote-traversable namespaces
    # (modules, packages, instances) must have an export list!
    if not hasattr(container, '__exports__'):
        raise errors.TraversalError(request.get_path(), 
            "%s has no __exports__ list" % repr(container))

    # Second security check: call __access__ function if it's present.
    if hasattr(container, '__access__'):
        # will raise AccessError if access failed
        container.__access__(request)
        
    # Third security check: make sure the current name component is
    # in the export list.
    if component in container.__exports__:
        is_module = type(container) is types.ModuleType
        has_component = hasattr(container, component)
        if is_module and not has_component:
            # get next module
            mod_name = container.__name__ + '.' + component
            # If we get an ImportError here we don't catch it.  It means
            # that either someone exported something that doesn't exist or
            # an exception was raised from deeping in the code.  A
            # traceback should be generated in either case.
            object = get_module(mod_name)

        elif has_component:
            object = getattr(container, component)

        else: 
            # not a module, doesn't have 'component'
            raise errors.TraversalError(request.get_path(),
                "object %s has no attribute '%s'" % (`container`, component))
    else:
        # Component is not in exports list.  Check if namespace provides a
        # '__getname__()' function, which translates an arbitrary string
        # into an object that we continue traversing.  This is very handy;
        # it lets you put user-space objects into your URL-space,
        # eliminating the need for digging ID strings out of a query, or
        # checking PATHINFO after Quixote's done with it.  But it is a
        # compromise to security: it opens up the traversal algorithm to
        # arbitrary names not listed in __exports__!
        if hasattr(container, "__getname__"):
            object = container.__getname__(request, component)
        else:
            raise errors.TraversalError(request.get_path(),
                ("%s not in __exports__ list for %s" %
                 (repr(component), container)))
    
    return object

def traverse_url (path, request):
    """traverse_url(path : string, request : HTTPRequest)
       -> (script_name: string, path_info : string, object : any)

    Perform traversal based on the provided path, starting at the root
    object.  It returns the script name and path info values for
    the arrived-at object, along with the object itself.

    It's expected that the final object is something callable like a
    function or a method; intermediate objects along the way will
    usually be packages or modules.

    To prevent crackers from writing URLs that traverse private
    objects, every package, module, or object along the way must have
    an __exports__ attribute containing a list of publicly visible
    names.  Not having an __exports__ attribute is an error, though
    having __exports__ be an empty list is OK (though useless).  If a
    component of the path isn't in __exports__, that also produces an
    error.
    """
    from quixote.config import config
    
    # replace repeated slashes with a single slash
    path = re.sub("//*", "/", path)

    # split path apart; /foo/bar/baz  -> ['foo', 'bar', 'baz']
    #                   /foo/bar/     -> ['foo', 'bar', '']
    path_components = string.split(path[1:], '/')

    # Traverse starting at the root
    object = get_module(config.root_namespace)

    # Loop over the components of the path
    index = 0
    for component in path_components:
        index = index + 1
        if component == "":
            # "/q/foo/" == "/q/foo/index"
            component = "index"
        object = get_component(object, component, path, request)

    if not callable(object):
        # We went through all the components of the path and ended up at
        # something which isn't callable, like a module or an instance
        # without a __call__ method.
        if path[-1] != '/':
            if not request.form and config.fix_trailing_slash:
                # This is for the convenience of users who type in paths.
                # Repair the path and redirect.  This should no happen for
                # URLs within the site.
                path = "%s/%s/" % (config.url_prefix,
                                   string.join(path_components, '/'))
                def redirect_slash(request, path=path):
                    return request.redirect(path, permanent=1)
                return '', '', redirect_slash
            else:
                # Automatic redirects disabled or there is form data.  If
                # there is form data then the programmer is using the
                # wrong path.  A redirect won't work if the form data came
                # from a POST anyhow.
                raise errors.TraversalError(path, "missing trailing slash?")
        else:
            raise error.TraversalError(path, "uncallable object %s" % `object`)

    # Determine the script name and path info to be returned.
    script = string.join([config.url_prefix] + path_components[:index], '/')
    path_info = ""

    return script, path_info, object


def mail_error (msg):
    """mail_error(msg : string) -> None

    If Quixote has been configured to do so, send an e-mail notifying
    someone of a traceback. 
    """
    from quixote.config import config
    if config.error_email is None: return

    cmd = "/usr/lib/sendmail -f Quixote '%s'" % config.error_email
    pipe = os.popen(cmd, 'w')
    pipe.write('Subject: Quixote Traceback\n')
    pipe.write('From: Quixote on %s <%s>\n' % (socket.gethostname(),
                                               config.error_email) )
    pipe.write('\n')
    pipe.write(msg)
    pipe.close()


def error_handler (request, exception, error_msg):
    """error_handler(request : HTTPRequest,
                     exception : Exception,
                     error_msg: string) -> string

    Return an error page to the user.  This function is called if the
    configuration variable ERROR_HANDLER is not set.  The error_msg
    contains a traceback and a dump if the request and response objects.  
    """
    from quixote.config import config
    request.response.setHeader("Content-Type", "text/plain")
    request.response.setStatus(500)
    if config.display_exceptions:
        return error_msg
    else:
        return ("An error occurred while handling your request.  Check "
                "the Quixote\nerror log for details.")

def report_error (request, original_response, exc_info):
    from quixote.config import config
    exc_type, exc_value, tb = exc_info
    error_file = cStringIO.StringIO()
    # format the traceback
    error_file.write('Traceback\n'
                     '=========\n')
    traceback.print_exception(exc_type, exc_value, tb, file=error_file)
    # include request and response dumps
    error_file.write('\nRequest\n'
                       '=======\n')
    error_file.write(request.dump())
    error_file.write('\nOriginal Response\n'
                       '=================\n')
    error_file.write(str(original_response))
    error_msg = error_file.getvalue()
    if config.error_log is None:
        sys.stderr.write(error_msg)
    else:
        try:
            errlog = open(config.error_log, 'a')
        except IOError, exc:
            sys.stderr.write("error opening error log %s: %s\n" 
                             % (`config.error_log`, exc.strerror))
        else:
            errlog.write(error_msg)
            errlog.close()
    mail_error(error_msg)
    # return an error message to the browser
    func = error_handler
    if config.error_handler:
        components = string.split(config.error_handler, ".")
        module_name = string.join(components[:-1], ".")
        func_name = components[-1]
        module = get_module(module_name)
        func = getattr(module, func_name, func)
    return func(request, exc_value, error_msg)
    

exit_now = 0
    
def call (stdin, stdout, stderr):
    """call(stdin : file, stdout : file, stderr : file)

    The master function that does all the work for a single request.
    call() uses traverse_url() to get a callable object, creates the
    request object, and calls the object.  If it raises an exception, the
    traceback is logged and e-mailed.

    Any extra arguments will be passed to the callable.
    """
    global exit_now

    sess_coll = session.get_session_collection()
    
    # Build request, response objects
    request = HTTPRequest.HTTPRequest(stdin, os.environ)
    request.processInputs()

    try:
        # Get the session object and stick it onto the request
        request.session = session.get_session(request)

        # Traverse package to a (hopefully-) callable object
        (os.environ['SCRIPT_NAME'], os.environ['PATH_INFO'], object) = \
            traverse_url(os.environ['PATH_INFO'], request)

        if not callable(object):
            raise RuntimeError, "object not callable: %s" % repr(object)

        # Call object
        try:
            output = object(request)
        except SystemExit:
            output = "Server restarted"
            exit_now = 1

        if output is None:
            raise RuntimeError, 'Callable %s returned None' % repr(object)

    except errors.PublishError, exc:
        # abort the publishing loop and return a result right way
        sess_coll.abort() # XXX Provide a way of controlling this!  Maybe
                          # the PublishError exception should have an
                          # attribute to control this.
        request.response = HTTPResponse.HTTPResponse()
        output = exc.format(request)
        
    except:
        # some other exception, generate error messages to the logs, etc.
        sess_coll.abort()
        # build new response to be safe
        old_response = request.response
        request.response = HTTPResponse.HTTPResponse()
        output = report_error(request, old_response, sys.exc_info())

    else:
        # The callable ran OK, commit any changes to the session
        sess_coll.commit()

    # Output results from Response object
    request.response.setBody(output)
    stdout.write(str(request.response))


def call_fcgi ():
    """call_fcgi() -> None

    Entry point from CGI scripts; it will repeatedly do the call()
    function until there are no more FCGI requests.  This function can
    also be called in regular CGI scripts, in which case call() is
    called exactly once.
    """
    from quixote.config import config
    while fcgi.isFCGI() and not exit_now:
        f = fcgi.FCGI()
        os.environ = f.env
        call(f.inp, f.out, f.err)
        f.Finish()
        if config.run_once: break
